import numpy as np
import numpy.testing as npt
from astropy.utils.data import get_pkg_data_filename
from astropy.tests.helper import raises

from skypy.utils.special import gammaincc


def test_gammaincc_scalar():
    # test with scalar input and expect scalar output
    npt.assert_allclose(gammaincc(1.2, 1.5), 0.28893139222051745)


def test_gammaincc_array():
    # test with vector a
    npt.assert_allclose(gammaincc([-1.2, 1.2], 1.5),
                        [0.008769092458747352, 0.28893139222051745])

    # test with vector x
    npt.assert_allclose(gammaincc(1.2, [0.5, 1.5]),
                        [0.6962998597584569, 0.28893139222051745])

    # test with vector a and x
    npt.assert_allclose(gammaincc([1.2, -1.2], [0.5, 1.5]),
                        [0.6962998597584569, 0.008769092458747352])

    # test with broadcast
    npt.assert_allclose(gammaincc([[1.2], [-1.2]], [0.5, 1.5]),
                        [[0.6962998597584569, 0.28893139222051745],
                         [0.1417443053403289, 0.008769092458747352]])


def test_gammaincc_edge_cases():
    # gammaincc is zero for x = inf
    assert gammaincc(1.2, np.inf) == 0
    assert gammaincc(-1.2, np.inf) == 0
    assert gammaincc(0, np.inf) == 0
    npt.assert_equal(gammaincc([1.2, 2.2], np.inf), [0, 0])
    npt.assert_equal(gammaincc([-1.2, -2.2], np.inf), [0, 0])
    npt.assert_equal(gammaincc([0.0, 1.0], np.inf), [0, 0])

    # gammaincc is zero for a = -1, -2, -3, ...
    assert gammaincc(-1.0, 0.5) == 0
    assert gammaincc(-2.0, 0.5) == 0
    npt.assert_equal(gammaincc([-1.0, -2.0], 0.5), [0, 0])

    # gammaincc is unity for a > 0 and x = 0
    assert gammaincc(0.5, 0) == 1
    assert gammaincc(1.5, 0) == 1
    npt.assert_equal(gammaincc([0.5, 1.5], 0), [1, 1])

    # gammaincc is zero for a nonpositive integer and x = 0
    assert gammaincc(0, 0) == 0
    assert gammaincc(-1, 0) == 0
    assert gammaincc(-2, 0) == 0
    npt.assert_equal(gammaincc([0, -1, -2], 0), [0, 0, 0])

    # gammaincc is infinity for a negative noninteger and x = 0
    assert gammaincc(-0.5, 0) == -np.inf
    assert gammaincc(-1.5, 0) == np.inf
    assert gammaincc(-2.5, 0) == -np.inf
    npt.assert_equal(gammaincc([-0.5, -1.5], 0), [-np.inf, np.inf])


def test_gammaincc_precision():
    # test precision against precomputed values
    values_file = get_pkg_data_filename('data/gammaincc.txt')
    a, x, v = np.loadtxt(values_file).T
    g = gammaincc(a, x)

    # collect where the assertion will fail before it does,
    # so we can have a more informative message
    fail = ~np.isclose(g, v, rtol=1e-10, atol=0)
    lines = []
    if np.any(fail):
        for numbers in zip(a[fail], x[fail], g[fail], v[fail]):
            lines.append('gammaincc(%g, %g) = %g != %g)' % numbers)

    # now do the assertion
    npt.assert_allclose(g, v, rtol=1e-10, atol=0, err_msg='\n'.join(lines))


@raises(ValueError)
def test_gammaincc_neg_x_scalar():
    # negative x raises an exception
    gammaincc(0.5, -1.0)


@raises(ValueError)
def test_gammaincc_neg_x_array():
    # negative x in array raises an exception
    gammaincc(0.5, [3.0, 2.0, 1.0, 0.0, -1.0])



"""Galaxy ellipticity module.

This module provides facilities to sample the ellipticities of galaxies.
"""

import numpy as np
from scipy import stats


__all__ = [
    'beta_ellipticity',
    'ryden04',
]


def beta_ellipticity(e_ratio, e_sum, size=None):
    r'''Galaxy ellipticities sampled from a reparameterized beta distribution.

    The ellipticities follow a beta distribution parameterized by
    :math:`e_{\rm ratio}` and :math:`e_{\rm sum}` as presented in [1]_ Section
    III.A.

    Parameters
    ----------
    e_ratio : array_like
        Mean ellipticity of the distribution, must be between 0 and 1.
    e_sum : array_like
        Parameter controlling the width of the distribution, must be positive.

    Notes
    -----
    The probability distribution function :math:`p(e)` for ellipticity
    :math:`e` is given by a beta distribution:

    .. math::

        p(e) \sim \frac{\Gamma(a+b)}{\Gamma(a) \Gamma(b)} x^{a-1} (1-x)^{b-1}

    for :math:`0 <= e <= 1`, :math:`a = e_{\rm sum} e_{\rm ratio}`,
    :math:`b = e_{\rm sum} (1 - e_{\rm ratio})`, :math:`0 < e_{\rm ratio} < 1`
    and :math:`e_{\rm sum} > 0`, where :math:`\Gamma` is the gamma function.

    References
    ----------
    .. [1] Kacprzak T., Herbel J., Nicola A. et al., arXiv:1906.01018

    Examples
    --------
    Sample 10000 random variates from the Kacprzak model with
    :math:`e_{\rm ratio} = 0.5` and :math:`e_{\rm sum} = 1.0`:

    >>> from skypy.galaxy.ellipticity import beta_ellipticity
    >>> ellipticity = beta_ellipticity(0.5, 1.0, size=10000)

    '''

    # convert to beta distribution parameters
    a = e_sum * e_ratio
    b = e_sum * (1.0 - e_ratio)

    # sample from the beta distribution
    return np.random.beta(a, b, size)


def ryden04(mu_gamma, sigma_gamma, mu, sigma, size=None):
    r'''Ellipticity distribution of Ryden (2004).

    The ellipticity is sampled by randomly projecting a 3D ellipsoid with
    principal axes :math:`A > B > C` [1]_. The distribution of the axis ratio
    :math:`\gamma = C/A` is a truncated normal with mean :math:`\mu_\gamma` and
    standard deviation :math:`\sigma_\gamma`. The distribution of
    :math:`\epsilon = \log(1 - B/A)` is truncated normal with mean :math:`\mu`
    and standard deviation :math:`\sigma`.

    Parameters
    ----------
    mu_gamma : array_like
        Mean of the truncated Gaussian for :math:`\gamma`.
    sigma_gamma : array_like
        Standard deviation for :math:`\gamma`.
    mu : array_like
        Mean of the truncated Gaussian for :math:`\epsilon`.
    sigma : array_like
        Standard deviation for :math:`\epsilon`.
    size : int or tuple of ints or None
        Size of the sample. If `None` the size is inferred from the parameters.

    References
    ----------
    .. [1] Ryden B. S., 2004, ApJ, 601, 214

    Examples
    --------
    Sample 10000 random variates from the Ryden (2004) model with parameters
    :math:`\mu_\gamma = 0.222`, :math:`\sigma_\gamma = 0.056`,
    :math:`\mu = -1.85`, and :math:`\sigma = 0.89`.

    >>> from skypy.galaxy.ellipticity import ryden04
    >>> ellipticity = ryden04(0.222, 0.056, -1.85, 0.89, size=10000)

    '''

    # get size if not given
    if size is None:
        size = np.broadcast(mu_gamma, sigma_gamma, mu, sigma).shape

    # truncation for gamma standard normal
    a_gam = np.divide(np.negative(mu_gamma), sigma_gamma)
    b_gam = np.divide(np.subtract(1, mu_gamma), sigma_gamma)

    # truncation for log(epsilon) standard normal
    a_eps = -np.inf
    b_eps = np.divide(np.negative(mu), sigma)

    # draw gamma and epsilon from truncated normal -- eq.s (10)-(11)
    gam = stats.truncnorm.rvs(a_gam, b_gam, mu_gamma, sigma_gamma, size=size)
    eps = np.exp(stats.truncnorm.rvs(a_eps, b_eps, mu, sigma, size=size))

    # scipy 1.5.x bug: make scalar if size is empty
    if size == () and not np.isscalar(gam):  # pragma: no cover
        gam, eps = gam.item(), eps.item()

    # draw random viewing angle (theta, phi)
    cos2_theta = np.random.uniform(low=-1., high=1., size=size)**2
    cos2_phi = np.cos(np.random.uniform(low=0., high=2*np.pi, size=size))**2
    sin2_theta = 1 - cos2_theta
    sin2_phi = 1 - cos2_phi

    # compute Binney's ABC -- eq.s (13)-(15)
    A = (1 - eps*(2-eps)*sin2_phi)*cos2_theta + gam**2*sin2_theta
    B = (2*eps*(2-eps))**2*cos2_theta*sin2_phi*cos2_phi
    C = 1 - eps*(2-eps)*cos2_phi

    # compute axis ratio q -- eq. (12)
    q = np.sqrt((A+C-np.sqrt((A-C)**2+B))/(A+C+np.sqrt((A-C)**2+B)))

    # return the ellipticity
    return (1-q)/(1+q)



"""Driver module.

This module provides methods to run pipelines of functions with dependencies
and handle their results.
"""

from copy import deepcopy
from importlib import import_module
import builtins
import networkx


__all__ = [
    'SkyPyDriver',
]


class SkyPyDriver:
    r'''Class for running pipelines.

    This is the main class for running pipelines of functions with dependencies
    and using their results to generate variables and tables.
    '''

    def execute(self, configuration, file_format=None, overwrite=False):
        r'''Run a pipeline.

        This function runs a pipeline of functions to generate variables and
        the columns of a set of tables. It uses a Directed Acyclic Graph to
        determine a non-blocking order of execution that resolves any
        dependencies, see [1]_. Tables can optionally be written to file.

        Parameters
        ----------
        configuration : dict-like
            Configuration for the pipeline.
        file_format : str
            File format used to write tables. Files are written using the
            Astropy unified file read/write interface; see [2]_ for supported
            file formats. If None (default) tables are not written to file.
        overwrite : bool
            Whether to overwrite any existing files without warning.

        Notes
        -----
        Each step in the pipeline is configured by a dictionary specifying
        a variable name and the associated value.

        A value that is a tuple `(function_name, function_args)` specifies that
        the value will be the result of a function call. The first item is the
        fully qualified function name, and the second value specifies the
        function arguments.

        If a function argument is a tuple `(variable_name,)`, it refers to the
        values of previous step in the pipeline. The tuple item must be the
        name of the reference variable.

        'configuration' should contain the name and configuration of each
        variable and/or an entry named 'tables'. 'tables' should contain a set
        of nested dictionaries, first containing the name of each table, then
        the name and configuration of each column and optionally an entry named
        'init' with a configuration that initialises the table. If 'init' is
        not specificed the table will be initialised as an empty astropy Table
        by default.

        See [3]_ for examples of pipeline configurations in YAML format.

        References
        ----------
        .. [1] https://networkx.github.io/documentation/stable/
        .. [2] https://docs.astropy.org/en/stable/io/unified.html
        .. [3] https://github.com/skypyproject/skypy/tree/master/examples
        '''

        # config contains settings for all variables and table initialisation
        # table_config contains settings for all table columns
        config = deepcopy(configuration)
        table_config = config.pop('tables', {})
        default_table = ('astropy.table.Table',)
        config.update({k: v.pop('.init', default_table)
                      for k, v in table_config.items()})

        # Create a Directed Acyclic Graph of all jobs and dependencies
        dag = networkx.DiGraph()

        # - add nodes for each variable, table and column
        # - add edges for the table dependencies
        # - keep track where functions need to be called
        #   functions are tuples (function name, [function args])
        functions = {}
        for job, settings in config.items():
            dag.add_node(job)
            if isinstance(settings, tuple):
                functions[job] = settings
        for table, columns in table_config.items():
            table_complete = '.'.join((table, 'complete'))
            dag.add_node(table_complete)
            dag.add_edge(table, table_complete)
            for column, settings in columns.items():
                job = '.'.join((table, column))
                dag.add_node(job)
                dag.add_edge(table, job)
                dag.add_edge(job, table_complete)
                if isinstance(settings, tuple):
                    functions[job] = settings

        # go through functions and add edges for all references
        for job, settings in functions.items():
            # settings are tuple (function, [args])
            args = settings[1] if len(settings) > 1 else None
            # get dependencies from arguments
            deps = self.get_deps(args)
            # add edges for dependencies
            for d in deps:
                if dag.has_node(d):
                    dag.add_edge(d, job)
                else:
                    raise KeyError(d)

        # Execute jobs in order that resolves dependencies
        for job in networkx.topological_sort(dag):
            if job.endswith('.complete'):
                continue
            elif job in config:
                settings = config.get(job)
                setattr(self, job, self.get_value(settings))
            else:
                table, column = job.split('.')
                settings = table_config[table][column]
                getattr(self, table)[column] = self.get_value(settings)

        # Write tables to file
        if file_format:
            for table in table_config.keys():
                filename = '.'.join((table, file_format))
                getattr(self, table).write(filename, overwrite=overwrite)

    def get_value(self, value):
        '''return the value of a field

        tuples specify function calls `(function name, function args)`
        '''

        # check for plain value
        if not isinstance(value, tuple):
            return value

        # value is tuple (function name, [function args])
        name = value[0]
        args = value[1] if len(value) > 1 else []

        # Import function
        function_path = name.split('.')
        module = builtins
        for i, key in enumerate(function_path[:-1]):
            if not hasattr(module, key):
                module_name = '.'.join(function_path[:i+1])
                try:
                    module = import_module(module_name)
                except ModuleNotFoundError:
                    raise ModuleNotFoundError(module_name)
            else:
                module = getattr(module, key)
        function = getattr(module, function_path[-1])

        # Parse arguments
        parsed_args = self.get_args(args)

        # Call function
        if isinstance(args, dict):
            result = function(**parsed_args)
        elif isinstance(args, list):
            result = function(*parsed_args)
        else:
            result = function(parsed_args)

        return result

    def get_args(self, args):
        '''parse function arguments

        tuples specify references to container: `(field name,)`
        '''

        if isinstance(args, tuple):
            # get reference
            return self[args[0]]
        elif isinstance(args, dict):
            # recurse kwargs
            return {k: self.get_args(v) for k, v in args.items()}
        elif isinstance(args, list):
            # recurse args
            return [self.get_args(a) for a in args]
        else:
            # return value
            return args

    def get_deps(self, args):
        '''get dependencies from function args

        returns a list of all dependencies found
        '''

        if isinstance(args, tuple):
            # reference
            return [args[0]]
        elif isinstance(args, dict):
            # get explicit dependencies
            deps = args.pop('.depends', [])
            # turn a single value into a list
            if isinstance(deps, str) or not isinstance(deps, list):
                deps = [deps]
            # recurse remaining kwargs
            return deps + sum([self.get_deps(a) for a in args.values()], [])
        elif isinstance(args, list):
            # recurse args
            return sum([self.get_deps(a) for a in args], [])
        else:
            # no reference
            return []

    def __getitem__(self, label):
        name, _, key = label.partition('.')
        item = getattr(self, name)
        return item[key] if key else item
